import os
from abc import ABCMeta, abstractproperty
import six


@six.add_metaclass(ABCMeta)
class DataFile(object):
    def __iter__(self):
        return self.iterator_class(self.content)

    def _is_gzip(self):
        return os.path.splitext(self.filename)[-1] == '.gz'

    @abstractproperty
    def filename(self):
        '''Retrieve filename'''

    @abstractproperty
    def key(self):
        '''Retrieve key (used in caching function)'''

    @abstractproperty
    def content(self):
        '''Content in data file.
        Compressed archives will be automatically decompressed.
        '''

    @property
    def raw_content(self):
        '''Raw content in data file'''



import hashlib

from akagi.data_source import DataSource
from akagi.data_file_bundles import S3DataFileBundle, LocalDataFileBundle
from akagi.iterator import FileFormat


class S3DataSource(DataSource):
    '''S3DataSource replesents a set of files on Amazon S3 bucket.
    '''

    @classmethod
    def for_prefix(cls, bucket_name, prefix, file_format=FileFormat.BINARY, no_cache=False):
        return S3DataSource(bucket_name, prefix=prefix, keys=None, file_format=file_format, no_cache=no_cache)

    @classmethod
    def for_keys(cls, bucket_name, keys, file_format=FileFormat.BINARY, no_cache=False):
        return S3DataSource(bucket_name, prefix=None, keys=keys, file_format=file_format, no_cache=no_cache)

    @classmethod
    def for_key(cls, bucket_name, key, file_format=FileFormat.BINARY, no_cache=False):
        return S3DataSource(bucket_name, prefix=None, keys=[key], file_format=file_format, no_cache=no_cache)

    def _setup_bundle(self):
        self.bundle = S3DataFileBundle(self._bucket_name, self._prefix, self._keys, self._file_format)

        if not self._no_cache:
            self.save()
            self.bundle = LocalDataFileBundle(self._local_cache_dir, self._file_format)

    def __init__(self, bucket_name, prefix=None, keys=None, file_format=FileFormat.BINARY, no_cache=False):
        self._file_format = file_format
        self._bucket_name = bucket_name
        self._keys = keys
        self._prefix = prefix
        self._no_cache = no_cache

        if self.exists_local() and not self._no_cache:
            self.bundle = LocalDataFileBundle(self._local_cache_dir, self._file_format)
        else:
            self.bundle = None

    def __iter__(self):
        if self.bundle is None:
            self._setup_bundle()

        return iter(self.bundle)

    @property
    def _hex_hash(self):
        sig = [self._bucket_name]
        if self._prefix:
            sig.append(self._prefix)

        if self._keys:
            sig.extend(self._keys)

        sig = '$$'.join(sig)
        return hashlib.sha256(sig.encode('utf-8')).hexdigest()



import json
import os

from akagi.iterator import Iterator, FileFormat
from akagi.data_file_bundle import DataFileBundle
from akagi.data_files import LocalDataFile


class LocalDataFileBundle(DataFileBundle):
    def __init__(self, directory_path, file_format=FileFormat.CSV):
        self.directory_path = os.path.expanduser(directory_path)
        self._file_format = file_format
        self._iterator_class = None

    @property
    def data_files(self):
        paths = []

        for root, _, filenames in os.walk(self.directory_path):
            for filename in filenames:
                paths.append(os.path.join(root, filename))

        return [LocalDataFile(path, self.iterator_class) for path in paths]

    @property
    def iterator_class(self):
        if self._iterator_class is None:
            if 'iterator_class' in self.metadata:
                self._iterator_class = Iterator.get_iterator_class(self._file_format)
            else:
                self._iterator_class = Iterator.get_iterator_class(self._file_format)

        return self._iterator_class

    @property
    def metadata(self):
        if os.path.isfile(self._metadata_path):
            with open(self._metadata_path, 'r') as f:
                return json.load(f)
        else:
            return {}

    @property
    def _metadata_path(self):
        return os.path.dirname(self.directory_path) + '.json'

    def clear(self):
        # XXX: LocalDataFileBundle won't do anything on clear for local files safety.
        pass



import csv
import hashlib
import os
import six

import MySQLdb

from akagi.data_source import DataSource
from akagi.data_file_bundles import LocalDataFileBundle
from akagi.log import logger


class MySQLDataSource(DataSource):
    '''MySQLDataSource replesents a set of row data as a result of query param.
    '''

    def __init__(self, query, db_conf={}, no_cache=False, keep_connection=False):
        self.query = query
        self.__db_conf = db_conf
        self.__connection = None
        self._keep_connection = keep_connection
        self._no_cache = no_cache

        if self.exists_local():
            self.bundle = LocalDataFileBundle(self._local_cache_dir)
        else:
            self.bundle = None

    @property
    def _connection(self):
        if self._keep_connection:
            if self.__connection is None:
                self.__connection = MySQLdb.connect(**self._db_conf)

            return self.__connection
        else:
            return MySQLdb.connect(**self._db_conf)

    def _setup_bundle(self):
        self.__result = []
        c = self._connection.cursor()
        logger.debug("Executing query...")
        c.execute(self.query)
        logger.debug("Finished.")

        rows = c.fetchall()

        tmp_csv_path = os.path.join('/tmp', 'akagi', self._hex_hash) + '.csv'
        tmp_csv_dir = os.path.dirname(tmp_csv_path)

        if not os.path.isdir(tmp_csv_dir):
            os.makedirs(tmp_csv_dir, mode=0o755)

        with open(tmp_csv_path, 'w') as f:
            csv.writer(f).writerows(rows)

            self.bundle = LocalDataFileBundle(tmp_csv_dir, file_format='csv')

        if not self._no_cache:
            self.save()
            self.bundle = LocalDataFileBundle(self._local_cache_dir)

    def __iter__(self):
        if self.bundle is None:
            self._setup_bundle()

        return iter(self.bundle)

    @property
    def _tmp_csv_path(self):
        return os.path.join('/tmp')

    @property
    def _db_conf(self):
        conf = {
            'host':
            self.__db_conf.get('host') or os.getenv('MYSQL_DB_HOST', 'localhost'),
            'user':
            self.__db_conf.get('user') or os.getenv('MYSQL_DB_USER'),
            'passwd':
            self.__db_conf.get('password') or os.getenv('MYSQL_DB_PASS'),
            'db':
            self.__db_conf.get('db') or os.getenv('MYSQL_DB_NAME'),
            'port':
            self.__db_conf.get('port') or os.getenv('MYSQL_DB_PORT', 3306),
            'unix_socket':
            self.__db_conf.get('unix_socket') or os.getenv('MYSQL_DB_SOCKET')}

        return {k: v for k, v in six.iteritems(conf) if v is not None}

    @property
    def _hex_hash(self):
        return hashlib.sha256(self.query.encode('utf-8')).hexdigest()



import re
import six
from six import BytesIO

import gzip


def gzip_compress(data):
    if six.PY2:
        out_io = BytesIO()
        gzip.GzipFile(fileobj=out_io, mode='wb').write(data.read())
        out_io.seek(0)
        return out_io.read()
    else:
        return gzip.compress(data.read())


def gzip_decompress(data):
    if six.PY2:
        in_io = BytesIO()
        in_io.write(data.read())
        in_io.seek(0)
        return gzip.GzipFile(fileobj=in_io, mode='rb').read()
    else:
        return gzip.decompress(data.read())


def normalize_path(path):
    if path is None:
        return None
    else:
        return re.sub(r'^/', '', re.sub(r'\/{2,}', '/', path))



import csv
from six import StringIO
import six
import io


class CSVIterator(object):
    def __init__(self, content, skip_errors=True):
        self.content = self.decode(content)
        self._skip_errors = skip_errors
        self._iterator = csv.reader(self.content, escapechar='\\')

    @classmethod
    def open_file(cls, path):
        return open(path, newline='')

    def decode(self, content):
        if six.PY2:
            return content
        else:
            if isinstance(content, io.TextIOBase):
                return content
            else:
                return StringIO(content.decode('utf-8'))

    def __next__(self):
        try:
            return next(self._iterator)
        except StopIteration as e:
            raise e
        except Exception as e:
            if self._skip_errors:
                return next(self._iterator)
            else:
                raise e



from abc import ABCMeta, abstractmethod, abstractproperty
import six


@six.add_metaclass(ABCMeta)
class DataFileBundle(object):
    '''DataFileBundle is an base class of all data file bundles
    '''

    @abstractproperty
    def data_files(self):
        '''Retrieve the data files associated to the bundle.'''

    def __iter__(self):
        for i, df in enumerate(self.data_files):
            # XXX: cache here (S3DataFile => LocalDataFile)

            it = iter(df)
            while True:
                try:
                    yield next(it)
                except StopIteration as e:
                    break

    def __enter__(self):
        return self

    def __exit__(self, *exc_type):
        self.clear()
        return False

    @abstractmethod
    def clear(self):
        '''Clear associated datas.'''



import re
import boto3

from datetime import datetime

from akagi.iterator import Iterator, FileFormat
from akagi.data_files.s3_data_file import S3DataFile
from akagi.data_file_bundle import DataFileBundle
from akagi.log import logger
from akagi.utils import normalize_path


class S3DataFileBundle(DataFileBundle):
    def __init__(self, bucket_name, prefix=None, keys=None, file_format=FileFormat.CSV):
        self.bucket_name = bucket_name
        self.prefix = normalize_path(prefix)
        self.keys = keys

        if self.prefix is None and self.keys is None:
            raise Exception("Either prefix or keys must be set.")

        self.file_format = file_format
        self.iterator_class = Iterator.get_iterator_class(file_format)

        self.__s3 = None

    @property
    def data_files(self):
        if self.prefix is not None:
            return [S3DataFile(obj, self.iterator_class)
                    for obj in self._bucket.objects.filter(Prefix=self.prefix)]
        else:
            return [S3DataFile(self._bucket.Object(key), self.iterator_class)
                    for key in self.keys]

    def clear(self):
        for obj in self._bucket.objects.filter(Prefix=self.prefix):
            logger.debug("Deleting intermediate object on s3: %(key)s" % ({"key": obj.key}))
            obj.delete()

    @property
    def _bucket(self):
        return self._s3.Bucket(self.bucket_name)

    @property
    def _s3(self):
        if self.__s3 is None:
            self.__s3 = boto3.resource('s3')

        return self.__s3
